# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from openlineage.common.provider.great_expectations.action import (
    OpenLineageValidationAction,
)

__all__ = [OpenLineageValidationAction]
